/*
 * Copyright 2017-present Open Networking Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.onosproject.fpcagent;

import com.google.common.collect.Sets;
import org.apache.felix.scr.annotations.Activate;
import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Deactivate;
import org.apache.felix.scr.annotations.Reference;
import org.apache.felix.scr.annotations.ReferenceCardinality;
import org.apache.felix.scr.annotations.Service;
import org.onosproject.config.DynamicConfigEvent;
import org.onosproject.config.DynamicConfigListener;
import org.onosproject.config.DynamicConfigService;
import org.onosproject.core.ApplicationId;
import org.onosproject.core.CoreService;
import org.onosproject.core.IdGenerator;
import org.onosproject.fpcagent.protocols.DpnNgicCommunicator;
import org.onosproject.fpcagent.protocols.DpnP4Communicator;
import org.onosproject.fpcagent.providers.NgicDpnDeviceListener;
import org.onosproject.fpcagent.providers.NgicDpnProviderService;
import org.onosproject.fpcagent.util.ConfigHelper;
import org.onosproject.fpcagent.util.eventstream.JettyServer;
import org.onosproject.fpcagent.util.eventstream.NbEventWorkerManager;
import org.onosproject.fpcagent.util.eventstream.ParseStream;
import org.onosproject.fpcagent.workers.ZmqSbPublisherManager;
import org.onosproject.fpcagent.workers.ZmqSbSubscriberManager;
import org.onosproject.net.config.ConfigFactory;
import org.onosproject.net.config.NetworkConfigEvent;
import org.onosproject.net.config.NetworkConfigListener;
import org.onosproject.net.config.NetworkConfigRegistry;
import org.onosproject.net.config.NetworkConfigService;
import org.onosproject.net.config.basics.SubjectFactories;
import org.onosproject.net.flow.FlowRuleService;
import org.onosproject.restconf.api.RestconfService;
import org.onosproject.yang.model.ModelConverter;
import org.onosproject.yang.model.RpcRegistry;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashSet;
import java.util.Optional;

import static com.google.common.base.Preconditions.checkNotNull;
import static org.onosproject.fpcagent.util.FpcUtil.FPC_APP_ID;

/**
 * Fpc Manager.
 */
@Component(immediate = true)
@Service
public class FpcManager implements FpcService {
    private static final Logger log = LoggerFactory.getLogger(FpcManager.class);

    private static final Class<FpcConfig> CONFIG_CLASS = FpcConfig.class;
    private final InternalNetworkConfigListener configListener =
            new InternalNetworkConfigListener();

    private final InternalConfigListener dynListener =
            new InternalConfigListener();

    /* Services */
    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private FpcRpcService fpcRpcService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private ModelConverter modelConverter;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private DynamicConfigService dynamicConfigService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private RpcRegistry rpcRegistry;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private NetworkConfigRegistry registry;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private NetworkConfigService configService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private CoreService coreService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private NgicDpnProviderService ngicDpnProviderService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private RestconfService restconfService;

    @Reference(cardinality = ReferenceCardinality.MANDATORY_UNARY)
    private FlowRuleService flowRuleService;

    /* Variables */
    private IdGenerator notificationIds;
    private FpcConfig fpcConfig;
    private boolean started = false;
    private HashSet<NgicDpnDeviceListener> listeners = Sets.newHashSet();

    /* Config */
    private ConfigFactory<ApplicationId, FpcConfig> fpcConfigConfigFactory =
            new ConfigFactory<ApplicationId, FpcConfig>(
                    SubjectFactories.APP_SUBJECT_FACTORY, FpcConfig.class, "fpcagent") {
                @Override
                public FpcConfig createConfig() {
                    return new FpcConfig();
                }
            };

    @Activate
    protected void activate() {
        ApplicationId applicationId = coreService.registerApplication(FPC_APP_ID);
        configService.addListener(configListener);
        registry.registerConfigFactory(fpcConfigConfigFactory);
        dynamicConfigService.addListener(dynListener);

        notificationIds = coreService.getIdGenerator("fpc-notification-ids");

        JettyServer.createInstance().open();
        ParseStream.createInstance().open();
        NbEventWorkerManager.createInstance(20).open();
        DpnNgicCommunicator.createInstance();
        DpnP4Communicator.createInstance(applicationId, flowRuleService);

        log.info("FPC Service Started");
    }

    @Deactivate
    protected void deactivate() {
        configService.removeListener(configListener);
        dynamicConfigService.removeListener(dynListener);
        registry.unregisterConfigFactory(fpcConfigConfigFactory);

        if (started) {
            ZmqSbSubscriberManager.getInstance().close();
            ZmqSbPublisherManager.getInstance().close();
            started = false;
        }

        NbEventWorkerManager.getInstance().close();
        ParseStream.getInstance().close();
        JettyServer.getInstance().close();

        log.info("FPC Service Stopped");
    }

    /**
     * Initialize ZMQ Managers based on configuration.
     */
    private void init() {
        fpcConfig.getConfig().ifPresent(
                helper -> {
                    started = true;
                    ZmqSbSubscriberManager.createInstance(
                            helper.dpnSubscriberUri(),
                            helper.nodeId(),
                            helper.networkId(),
                            ngicDpnProviderService.getListener(),
                            notificationIds
                    );

                    ZmqSbPublisherManager.createInstance(
                            helper.dpnPublisherUri(),
                            helper.dpnClientThreads()
                    );

                    ZmqSbPublisherManager.getInstance().open();
                    ZmqSbSubscriberManager.getInstance().open();
                }
        );
    }

    @Override
    public void addListener(NgicDpnDeviceListener listener) {
        listeners.add(listener);
    }

    @Override
    public void removeListener(NgicDpnDeviceListener listener) {
        listeners.remove(listener);
    }

    @Override
    public Optional<ConfigHelper> getConfig() {
        return fpcConfig != null ? fpcConfig.getConfig() : Optional.empty();
    }

    private class InternalNetworkConfigListener implements NetworkConfigListener {

        @Override
        public void event(NetworkConfigEvent event) {
            switch (event.type()) {
                case CONFIG_REGISTERED:
                case CONFIG_UNREGISTERED: {
                    break;
                }
                case CONFIG_REMOVED: {
                    if (event.configClass() == CONFIG_CLASS) {
                        fpcConfig = null;
                    }
                    break;
                }
                case CONFIG_UPDATED:
                case CONFIG_ADDED: {
                    if (event.configClass() == CONFIG_CLASS) {
                        event.config().ifPresent(config -> {
                            fpcConfig = (FpcConfig) config;
                            init();
                        });
                    }
                    break;
                }
                default:
                    break;
            }
        }

    }

    /**
     * Representation of internal listener, listening for dynamic config event.
     */
    private class InternalConfigListener implements DynamicConfigListener {

        @Override
        public void event(DynamicConfigEvent event) {
            checkNotNull(event, "Event cannot be NULL");
            log.debug("event {}", event);
        }
    }
}
