/*
 * Copyright 2017-present Open Networking Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.onosproject.fpcagent.workers;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.onosproject.core.IdGenerator;
import org.onosproject.fpcagent.providers.NgicDpnDeviceListener;
import org.onosproject.fpcagent.util.CacheManager;
import org.onosproject.fpcagent.util.FpcUtil;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.ClientIdentifier;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.DefaultYangAutoPrefixNotify;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.NotificationId;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.OpIdentifier;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.yangautoprefixnotify
        .value.DefaultDownlinkDataNotification;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.yangautoprefixnotify
        .value.DownlinkDataNotification;
import org.onosproject.yang.gen.v1.ietfdmmfpcbase.rev20160803.ietfdmmfpcbase.FpcIdentity;
import org.onosproject.yang.gen.v1.ietfdmmfpcbase.rev20160803.ietfdmmfpcbase.fpcidentity.FpcIdentityUnion;
import org.onosproject.yang.model.ModelObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zeromq.ZContext;
import org.zeromq.ZMQ;

import java.math.BigInteger;
import java.util.AbstractMap;
import java.util.Arrays;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadLocalRandom;

import static org.onosproject.fpcagent.protocols.DpnNgicCommunicator.*;
import static org.onosproject.fpcagent.util.Converter.*;
import static org.onosproject.fpcagent.util.FpcUtil.createConfigResultNotification;
import static org.onosproject.fpcagent.util.FpcUtil.sendNotification;

public class ZmqSbSubscriberManager implements AutoCloseable {
    private static final Logger log = LoggerFactory.getLogger(ZmqSbSubscriberManager.class);
    private static final int MIN_TOPIC_VAL = 4;
    private static final int MAX_TOPIC_VAL = 255;
    private static ZmqSbSubscriberManager _instance = null;
    private final String address;
    private final String nodeId;
    private final String networkId;
    private Long controllerSourceId;
    private byte controllerTopic;
    private boolean run;
    private boolean conflictingTopic;
    private IdGenerator notificationIds;

    private Future<?> broadcastAllWorker;
    private Future<?> broadcastControllersWorker;
    private Future<?> broadcastTopicWorker;
    private Future<?> generalWorker;

    private NgicDpnDeviceListener ngicDpnDeviceListener;

    protected ZmqSbSubscriberManager(String address, String nodeId, String networkId,
                                     NgicDpnDeviceListener ngicDpnDeviceListener, IdGenerator notificationIds) {
        this.address = address;
        this.run = true;
        this.nodeId = nodeId;
        this.networkId = networkId;
        this.conflictingTopic = false;
        this.controllerSourceId = (long) ThreadLocalRandom.current().nextInt(0, 65535);
        this.ngicDpnDeviceListener = ngicDpnDeviceListener;
        this.notificationIds = notificationIds;
    }

    public static ZmqSbSubscriberManager createInstance(String address, String nodeId, String networkId,
                                                        NgicDpnDeviceListener providerService,
                                                        IdGenerator notificationIds) {
        if (_instance == null) {
            _instance = new ZmqSbSubscriberManager(address, nodeId, networkId, providerService, notificationIds);
        }
        return _instance;
    }

    public static ZmqSbSubscriberManager getInstance() {
        return _instance;
    }

    public byte getControllerTopic() {
        return controllerTopic;
    }

    public Long getControllerSourceId() {
        return controllerSourceId;
    }

    public void open() {
        broadcastAllWorker = Executors.newSingleThreadExecutor()
                .submit(new ZmqSubscriberWorker(ReservedTopics.BROADCAST_ALL.getType()));

        broadcastControllersWorker = Executors.newSingleThreadExecutor()
                .submit(new ZmqSubscriberWorker(ReservedTopics.BROADCAST_CONTROLLERS.getType()));

        broadcastTopicWorker = Executors.newSingleThreadExecutor()
                .submit(new AssignTopic());
    }

    @Override
    public void close() {
        sendGoodbyeDpns(nodeId, networkId);
        run = false;
    }

    /**
     * Interrupts the BroadcastTopicworker if there is an Assign topic Conflict.
     *
     * @param conflict - Flag to indicate conflict
     * @param subId    - Topic Id that caused the conflict
     */
    protected void broadcastAllSubIdCallBack(boolean conflict, byte subId) {
        if (conflict && controllerTopic == subId) {
            this.conflictingTopic = true;
            broadcastTopicWorker.cancel(true);
        }
    }

    /**
     * Broadcasts an Assign Conflict message.
     *
     * @param contents - byte array received over the southbound.
     */
    protected void sendAssignConflictMessage(byte[] contents) {
        byte topic = contents[2];
        short nodeIdLen = contents[7];
        short networkIdLen = contents[8 + nodeIdLen];
        String nodeId = new String(Arrays.copyOfRange(contents, 8, 8 + nodeIdLen));
        String networkId = new String(
                Arrays.copyOfRange(contents, 9 + nodeIdLen, 9 + nodeIdLen + networkIdLen)
        );

        if (controllerTopic == topic || (this.nodeId.equals(nodeId) && this.networkId.equals(networkId))) {
            sendAssignConflict(this.nodeId, this.networkId);
        }
    }

    protected class ZmqSubscriberWorker implements Runnable {
        private final byte subscribedTopic;
        private ZContext ctx;

        ZmqSubscriberWorker(byte subscribedTopic) {
            this.subscribedTopic = subscribedTopic;
            this.ctx = new ZContext();
        }

        /**
         * Ensures the session id is an unsigned 64 bit integer.
         *
         * @param sessionId - session id received from the DPN
         * @return unsigned session id
         */
        private BigInteger checkSessionId(BigInteger sessionId) {
            if (sessionId.compareTo(BigInteger.ZERO) < 0) {
                sessionId = sessionId.add(BigInteger.ONE.shiftLeft(64));
            }
            return sessionId;
        }

        /**
         * Decodes a DownlinkDataNotification.
         *
         * @param buf - message buffer
         * @param key - Concatenation of node id + / + network id
         * @return DownlinkDataNotification or null if it could not be successfully decoded
         */
        private DownlinkDataNotification processDdn(byte[] buf, String key) {
            DownlinkDataNotification ddn = new DefaultDownlinkDataNotification();
            ddn.sessionId(checkSessionId(toBigInt(buf, 2)));
            ddn.notificationMessageType("Downlink-Data-Notification");
            ddn.clientId(ClientIdentifier.of(FpcIdentity.of(FpcIdentityUnion.of(fromIntToLong(buf, 10)))));
            ddn.opId(OpIdentifier.of(BigInteger.valueOf(fromIntToLong(buf, 13))));
            CacheManager.getCaches().forEach(
                    cacheManager -> {
                        try {
                            cacheManager.nodeNetworkCache.get(key)
                                    .ifPresent(dpn -> ddn.notificationDpnId(dpn.dpnId()));
                        } catch (ExecutionException e) {
                            log.error(ExceptionUtils.getFullStackTrace(e));
                        }
                    }
            );
            return ddn;
        }

        public Map.Entry<Object, Object> decode(byte[] buf) {
//            uint8_t topic_id;
//            uint8_t type;
            S11MsgType type;
            type = S11MsgType.getEnum(buf[1]);
            if (type.equals(S11MsgType.DDN)) {
//                uint64_t session_id;
//                uint32_t client_id;
//                uint32_t op_id;
//                uint8_t  node_network_id_buffer[NN_ID_BUF_LEN];
                short nodeIdLen = buf[18];
                short networkIdLen = buf[19 + nodeIdLen];
                String key = new String(Arrays.copyOfRange(buf, 19, 19 + nodeIdLen))
                        + "/" + new String(
                                Arrays.copyOfRange(buf, 20 + nodeIdLen, 20 + nodeIdLen + networkIdLen)
                );
                return new AbstractMap.SimpleEntry<>(processDdn(buf, key), null);
            } else if (type.equals(S11MsgType.DPN_STATUS_INDICATION)) {
//                uint8_t  source_topic_id;
//                uint8_t  status;
//                uint32_t source;
//                uint8_t  node_network_id_buffer[NN_ID_BUF_LEN];
                DpnStatusIndication status;
                short nodeIdLen = buf[8];
                short networkIdLen = buf[9 + nodeIdLen];
                String deviceId = new String(Arrays.copyOfRange(buf, 9, 9 + nodeIdLen)) + "/"
                        + new String(Arrays.copyOfRange(buf, 10 + nodeIdLen, 10 + nodeIdLen + networkIdLen));
                status = DpnStatusIndication.getEnum(buf[3]);
                log.info("topic {}, status {}, source {}, node/net {}", (short) buf[2], status,
                        fromIntToLong(buf, 4), deviceId);
                if (status.equals(DpnStatusIndication.HELLO)) {
                    log.info("Hello {} on topic {}", deviceId, buf[2]);
                    ngicDpnDeviceListener.deviceAdded(deviceId, buf[2]);
                } else if (status.equals(DpnStatusIndication.GOODBYE)) {
                    log.info("Bye {}", deviceId);
                    ngicDpnDeviceListener.deviceRemoved(deviceId);
                }
                return new AbstractMap.SimpleEntry<>(status, deviceId);
            } else if (type.equals(S11MsgType.DPN_RESPONSE)) {
//                uint8_t  cause;
//                uint32_t client_id;
//                uint32_t op_id;
                ClientIdentifier clientId = ClientIdentifier.of(
                        FpcIdentity.of(FpcIdentityUnion.of(fromIntToLong(buf, 3)))
                );
                OpIdentifier opId = OpIdentifier.of(BigInteger.valueOf(fromIntToLong(buf, 7)));
                log.info("clientId {}, opId {}, cause {}", clientId.toString(), opId.toString(), (short) buf[2]);
                ModelObject notification = createConfigResultNotification(opId, buf[2]);
                sendNotification(clientId.toString(), notification);
            }
            return null;
        }

        @Override
        public void run() {
            ZMQ.Socket subscriber = this.ctx.createSocket(ZMQ.SUB);
            subscriber.connect(address);
            subscriber.subscribe(new byte[]{subscribedTopic});
            log.debug("Subscriber at {} / {}", address, subscribedTopic);
            while ((!Thread.currentThread().isInterrupted()) && run) {
                byte[] contents = subscriber.recv();
                byte topic = contents[0];
                S11MsgType messageType = S11MsgType.getEnum(contents[1]);
                log.info("Received {}", messageType);
                switch (topic) {
                    case 1:
                        if (messageType.equals(S11MsgType.ASSIGN_CONFLICT) &&
                                toInt(contents, 3) != controllerSourceId) {
                            broadcastAllSubIdCallBack(true, contents[2]);
                        } else if (messageType.equals(S11MsgType.ASSIGN_TOPIC) &&
                                toInt(contents, 3) != controllerSourceId) {
                            sendAssignConflictMessage(contents);
                        }
                        break;
                    default:
                        Map.Entry msg = decode(contents);
                        if (msg != null) {
                            Object key = msg.getKey();
                            if (key instanceof DownlinkDataNotification) {
                                DefaultYangAutoPrefixNotify notify = new DefaultYangAutoPrefixNotify();
                                notify.notificationId(NotificationId.of(notificationIds.getNewId()));
                                notify.timestamp(BigInteger.valueOf(System.currentTimeMillis()));
                                notify.value((DownlinkDataNotification) key);

                                sendNotification(((DownlinkDataNotification) key).clientId().toString(), notify);
                            } else if (key instanceof DpnStatusIndication) {
                                if (key.equals(DpnStatusIndication.HELLO)) {
                                    byte dpnTopic = FpcUtil.getTopicFromNode(msg.getValue().toString());
                                    if (dpnTopic != -1) {
                                        sendStatusAck(nodeId, networkId, dpnTopic);
                                    }
                                }
                            }
                        }
                }
            }
            subscriber.disconnect(address);
            subscriber.close();
            ctx.destroySocket(subscriber);
            try {
                Thread.currentThread().join();
            } catch (InterruptedException e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
            }
        }
    }

    /**
     * Class to broadcast a topic for the controller.
     */
    protected class AssignTopic implements Runnable {
        private byte topic;

        public AssignTopic() {
            this.topic = (byte) ThreadLocalRandom.current().nextInt(MIN_TOPIC_VAL, MAX_TOPIC_VAL + 1);
        }

        @Override
        public void run() {
            try {
                sendAssignTopic(nodeId, networkId, this.topic);
                log.debug("Thread sleeping: " + Thread.currentThread().getName());
                Thread.sleep(2000); // wait 10 sec before assigning topic
            } catch (InterruptedException e) {
                if (conflictingTopic) {
                    conflictingTopic = false;
                    this.topic = (byte) ThreadLocalRandom.current().nextInt(MIN_TOPIC_VAL, MAX_TOPIC_VAL + 1);
                    controllerTopic = this.topic;
                    this.run();
                    return;
                } else {
                    log.error(ExceptionUtils.getFullStackTrace(e));
                }
            }
            controllerTopic = this.topic;
            generalWorker = Executors.newSingleThreadExecutor().submit(new ZmqSubscriberWorker(this.topic));

            try {
                Thread.sleep(2000);
            } catch (InterruptedException e) {
                log.error(ExceptionUtils.getFullStackTrace(e));
            }

            sendHelloDpns(nodeId, networkId);
        }

    }
}
